def addition():
    """
    Turn the sequence of numbers into a sum

    + or - ?
    """


def multiply():
    """

    * or /

    """


def whatsbigger():
    """
    Given two numbers, determine which one bigger
    """


earef



import torch
from torch import nn, Tensor
from einops import rearrange
import torch.nn.functional as F
import torch

class LayerNorm(nn.Module):
    def __init__(self, d_model: int, bias: bool = False):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(d_model))
        self.bias = nn.Parameter(torch.zeros(d_model)) if bias else None

    def forward(self, x: Tensor) -> Tensor:
        return F.layer_norm(x, self.weight.shape, self.weight, self.bias, eps=1e-5)


class MLP(nn.Module):
    def __init__(self, d_model: int, bias: bool, dropout: float, mlp_factor=4):
        super().__init__()
        self.c_fc = nn.Linear(d_model, mlp_factor * d_model, bias=bias)
        self.c_proj = nn.Linear(mlp_factor * d_model, d_model, bias=bias)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: Tensor):
        x = self.c_fc(x)
        x = F.gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x


class Attention(nn.Module):
    """
    Memory Efficient or Flash Attention
    """

    def __init__(
        self,
        dmodel: int = 1024,
        n_heads: int = 16,
        bias: bool = False,
        dropout: float = 0.0,
        causal: bool = True,
        flash: bool = True,
    ):
        super().__init__()
        assert dmodel % n_heads == 0
        self.qkv = nn.Linear(dmodel, 3 * dmodel, bias=bias)
        self.c_proj = nn.Linear(dmodel, dmodel, bias=bias)
        self.attn_dropout = nn.Dropout(dropout)
        self.resid_dropout = nn.Dropout(dropout)
        self.n_head = n_heads
        self.n_embd = dmodel
        self.flash = flash
        self.dropout = dropout
        self.causal = causal

    def forward(self, x, mask=None):
        B, T, C = x.size()

        einop = "B T (split heads hs) -> split B heads T hs"
        q, k, v = rearrange(self.qkv(x), einop, heads=self.n_head, split=3)

        y = torch.nn.functional.scaled_dot_product_attention(
            q,
            k,
            v,
            dropout_p=self.dropout if self.training else 0,
            is_causal=self.causal,
            attn_mask=mask,
        )

        y = y.transpose(1, 2).contiguous().view(B, T, C)
        y = self.resid_dropout(self.c_proj(y))
        return y


class TransformerBlock(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_heads: int,
        bias: bool,
        dropout: float,
        causal: bool = True,
        mlp_factor = 2.0,
    ):
        super().__init__()
        self.attn_norm = LayerNorm(d_model, bias=bias)
        self.attn = Attention(d_model, n_heads, bias, dropout, causal)
        self.mlp_norm = LayerNorm(d_model, bias=bias)
        self.mlp = MLP(d_model, bias, dropout, mlp_factor)

    def forward(
        self,
        x: Tensor,
        mask: Tensor = None,
    ) -> Tensor:
        x = x + self.attn(self.attn_norm(x), mask=mask)
        x = x + self.mlp(self.mlp_norm(x))
        return x


class Decoder(nn.Module):
    def __init__(
        self,
        *,
        d_model: int,
        n_heads: int,
        n_layers: int,
        bias: bool,
        causal: bool = True,
        mlp_factor: float = 2,
        dropout: float = 0.0,
    ):
        super().__init__()

        self.drop = nn.Dropout(dropout)
        self.blocks = nn.ModuleList(
            [
                TransformerBlock(d_model, n_heads, bias, dropout, causal=causal, mlp_factor=mlp_factor)
                for _ in range(n_layers)
            ]
        )
        self.norm_f = LayerNorm(d_model, bias=bias)

    def forward(
        self,
        x: Tensor,
        mask: Tensor = None,
    ):
        x = self.drop(x)

        for block in self.blocks:
            x = block(x, mask)

        return self.norm_f(x)


class ScaledSinusoidalEmbedding(nn.Module):
    def __init__(self, d_model: int, theta=10000, max_seqlen: int = 8192):
        super().__init__()
        assert (d_model % 2) == 0
        self.weight = nn.Parameter(torch.ones(1) * d_model**-0.5)

        half_dim = d_model // 2
        freq_seq = torch.arange(half_dim).float() / half_dim
        inv_freq = theta**-freq_seq

        pos = torch.arange(max_seqlen)
        emb = torch.einsum("i, j -> i j", pos, inv_freq)
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)

        self.register_buffer("emb", emb, persistent=False)

    def forward(self, pos_ids: Tensor):
        emb = self.emb[pos_ids]
        return self.weight * emb



from torch.utils.data import IterableDataset
import random

class ToyDataset(IterableDataset):
    """
    Produces an infinite stream of toy tasks for training debugging
    """

    def __init__(self, challenges:list[callable], probs:list[float]=None):
        self.challenges = challenges
        self.probs = probs

        if self.probs is None:
           self.probs = [1/len(challenges)]*len(challenges)

        assert sum(self.probs) == 1

    def gen(self):
        while True:
            func = random.choices(self.challenges, self.probs)[0]
            yield func.generate(), func

    def __iter__(self):
        return self.gen()





import os
import time

import yaml
from pydantic import BaseModel

from toyds.constants import RUN_DIR
from toyds.utils import get_git_commit, get_git_branch, dictdiff, dict_to_strs
from toyds.download import dl_http_file


class ModelConfig(BaseModel):
    layers: int = 8
    dmodel: int = 1024
    nheads: int = 16
    dropout: float = 0
    bias: bool = False
    num_embs: int = 512
    max_seq_len: int = 1024


class DataConfig(BaseModel):
    num_workers: int = 4


class FlashConfig(BaseModel):
    enable_flash: bool = False
    enable_math: bool = True
    enable_mem_efficient: bool = True


class TrainConfig(BaseModel):
    gpus: int = 0
    total_steps: int = 1_000_000

    # If this is set to none, then the batch size will be determine automatically
    batch_size: int = 2
    # How many steps to do the forward before computing backward
    grad_acc_steps: int = 21
    flash: FlashConfig = FlashConfig()

    # AdamW
    min_lr: float = 1e-5
    max_lr: float = 5e-5
    weight_decay: float = 0.00
    betas: list[float] = [0.9, 0.95]  # YAML and tuples...

    max_grad_norm: float = 1.0
    log_batch_grad: float = 10.0
    log_batch_grad_steps: int = 1000

    lr_schedule: str = "warmup_then_cosine_decay"
    lr_warmup_steps: int = 800

    do_val: bool = False
    val_every: int = 10_000

    checkpoint_every: int = 2000
    checkpoint: str | None = None
    continue_from_checkpoint: bool = True

    amp_dtype: str | None = None
    torch_profile: bool = False
    wandb_watch: bool = False
    log_every: int = 1
    watch_every: int = 1000

    regression_win: int = 400
    regression_smoothing: float = 0.95

    @property
    def is_resuming(self) -> bool:
        return self.checkpoint is not None


class Config(BaseModel):
    data: DataConfig = DataConfig()
    model: ModelConfig = ModelConfig()
    train: TrainConfig = TrainConfig()
    seed: int = 42

    commit: str | None = None
    branch: str | None = None
    run_dir: str | None = None


def load_config(config: str | Config) -> Config:
    assert (
        os.getenv("WANDB_API_KEY") is not None
    ), "Please make sure you have set your `WANDB_API_KEY`"

    if type(config) is str:
        if "http" in config:
            config = dl_http_file(config)
        with open(config, encoding="utf-8") as f:
            config = Config(**yaml.safe_load(f))
    else:
        config = Config()

    if config.branch is None and config.commit is None:
        config.branch = get_git_branch()
        config.commit = get_git_commit()

    config.run_dir = os.path.join(RUN_DIR, str(int(time.time())))
    os.makedirs(config.run_dir, exist_ok=True)

    return config


def diffstr(config: Config, config2: Config) -> str:
    diff = dictdiff(config.model_dump(), config2.model_dump())
    result = dict_to_strs(diff)
    return result



import hashlib
import random
import subprocess
from typing import Union

import math

import numpy as np
import torch
import torch.distributed as dist
import hashlib
import os
import pickle
from functools import wraps
import torch.nn as nn
from torch import Tensor


def count_parameters(model: nn.Module, nongrad=False):
    return sum(p.numel() for p in model.parameters() if nongrad or p.requires_grad)


def get_git_commit() -> str:
    """
    https://stackoverflow.com/questions/14989858/get-the-current-git-hash-in-a-python-script
    """
    try:
        commit = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()
    except subprocess.CalledProcessError:
        commit = "unknown"

    return commit


def get_git_repo() -> str:
    try:
        repo = (
            subprocess.check_output(["git", "remote", "get-url", "origin"])
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        repo = "unknown"

    return repo


def get_git_branch() -> str:
    try:
        branch = (
            subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        branch = "unknown"

    return branch


def might_have_uncommitted_changes():
    try:
        msg = subprocess.check_output(["git", "status", "-s"]).decode().strip()
    except subprocess.CalledProcessError:
        msg = ""

    return len(msg) > 0


def reduce_tensor(tensor, world_size):
    rt = tensor.clone()
    dist.all_reduce(rt, op=dist.ReduceOp.SUM)
    if rt.is_floating_point():
        rt = rt / world_size
    else:
        rt = rt // world_size
    return rt


def seed_all(seed: int):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def get_random_states():
    # Get the state of torch, numpy and random libraries random state
    torch_state = torch.get_rng_state()
    np_state = np.random.get_state()
    random_state = random.getstate()
    return torch_state, np_state, random_state


def set_random_states(torch_state, np_state, random_state):
    """Set the state of the libraries"""
    torch.set_rng_state(torch_state)
    np.random.set_state(np_state)
    random.setstate(random_state)


def make_padding_mask(lengths: Tensor) -> Tensor:
    T_max = lengths.max()
    B = lengths.size(0)

    expanded_lengths = torch.arange(T_max).expand(B, T_max).to(lengths)

    return expanded_lengths >= lengths.unsqueeze(1)


def count_nans(x: Tensor):
    nan_mask = torch.isnan(x)
    num_nans = torch.sum(nan_mask).item()
    return num_nans


def prob_mask_like(shape, prob: float, device):
    if prob == 1:
        return torch.ones(shape, device=device, dtype=torch.bool)
    elif prob == 0:
        return torch.zeros(shape, device=device, dtype=torch.bool)
    else:
        return torch.zeros(shape, device=device).float().uniform_(0, 1) < prob


def multinomial(input: Tensor, num_samples: int, replacement=False, *, generator=None):
    """torch.multinomial with arbitrary number of dimensions, and number of candidates on the last dimension.

    Args:
        input (Tensor): The input tensor containing probabilities.
        num_samples (int): Number of samples to draw.
        replacement (bool): Whether to draw with replacement or not.
    Keywords args:
        generator (torch.Generator): A pseudorandom number generator for sampling.
    Returns:
        Tensor: Last dimension contains num_samples indices
            sampled from the multinomial probability distribution
            located in the last dimension of tensor input.
    """
    input_ = input.reshape(-1, input.shape[-1])
    output_ = torch.multinomial(
        input_, num_samples=num_samples, replacement=replacement, generator=generator
    )
    output = output_.reshape(*list(input.shape[:-1]), -1)
    return output


def apply_repetition_penalty(
    prev_ids: Tensor, next_logits: Tensor, repetition_penalty: float = 1.0
):
    score = torch.gather(next_logits, -1, prev_ids)
    score = torch.where(
        score < 0, score * repetition_penalty, score / repetition_penalty
    )
    next_logits = torch.scatter(next_logits, -1, prev_ids, score)
    return next_logits


def sample_top_k(probs: Tensor, k: int) -> Tensor:
    """Sample next token from top K values along the last dimension of the input probs tensor.

    Args:
        probs (Tensor): Input probabilities with token candidates on the last dimension.
        k (int): The k in “top-k”.
    Returns:
        Tensor: Sampled tokens.
    """
    top_k_value, _ = torch.topk(probs, k, dim=-1)
    min_value_top_k = top_k_value[..., [-1]]
    probs *= (probs >= min_value_top_k).float()
    probs.div_(probs.sum(dim=-1, keepdim=True))
    next_token = multinomial(probs, num_samples=1)
    return next_token


def sample_top_p(probs: Tensor, p: float) -> Tensor:
    """Sample next token from top P probabilities along the last dimension of the input probs tensor.

    Args:
        probs (Tensor): Input probabilities with token candidates on the last dimension.
        p (int): The p in “top-p”.
    Returns:
        Tensor: Sampled tokens.
    """
    probs_sort, probs_idx = torch.sort(probs, dim=-1, descending=True)
    probs_sum = torch.cumsum(probs_sort, dim=-1)
    mask = probs_sum - probs_sort > p
    probs_sort *= (~mask).float()
    probs_sort.div_(probs_sort.sum(dim=-1, keepdim=True))
    next_token = multinomial(probs_sort, num_samples=1)
    next_token = torch.gather(probs_idx, -1, next_token)
    return next_token


def sha256(b: Union[float, list, Tensor, str, bytes, np.ndarray]):
    if isinstance(b, (int, list, float)):
        b = str(b)
    if isinstance(b, Tensor):
        b = b.cpu().numpy()
    if isinstance(b, np.ndarray):
        b = b.tostring()
    if type(b) == str:
        b = b.encode()
    if type(b) == bytes:
        return hashlib.sha256(b).hexdigest()
    else:
        raise Exception("Not implemented a method to handle {0}".format(type(b)))

def to_device(obj: [nn.Module, Tensor, list, dict], targets: str | list[str]):
    """
    Takes obj and iterates through the keys putting them on the `device`
    """
    if isinstance(obj, Tensor) or isinstance(obj, nn.Module):
        return obj.to(targets)
    elif isinstance(obj, (list, tuple)):
        return [to_device(item, targets) for item in obj]
    elif isinstance(obj, dict):
        return {key: to_device(value, targets) for key, value in obj.items()}
    else:
        return obj


def hash_arguments(args, kwargs):
    arguments = list(args) + list(kwargs.keys()) + list(kwargs.values())
    return "".join([sha256(b) for b in arguments])

def cache(location=".cache") -> callable:
    def inner_function(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            os.makedirs(location, exist_ok=True)
            s = hash_arguments(args, kwargs)
            key = f.__name__ + s
            # Hash the args correctly
            fname = sha256(key)
            fname = os.path.join(location, fname)
            if os.path.exists(fname):
                with open(fname, "rb") as fl:
                    return pickle.load(fl)
            ret = f(*args, **kwargs)
            with open(fname, "wb") as fl:
                pickle.dump(ret, fl)
            return ret

        return wrapper

    return inner_function


def load_what_you_can(checkpoint: dict, model: nn.Module):
    """
    This method takes a checkpoint and loads as many weights from it as possible:

    If they are the same shape, there's nothing to do

    Will load the smallest shape otherwise.
    """
    model_state_dict = model.state_dict()
    checkpoint_state_dict = checkpoint

    for name, param in checkpoint_state_dict.items():
        if name not in model_state_dict:
            print(f"Ignoring parameter '{name}' because it is not found in the model")
            continue

        model_state = model_state_dict[name]
        mshape = model_state.shape
        pshape = param.shape

        if pshape == mshape:
            model_state.copy_(param)
            continue

        if len(pshape) != len(mshape):
            # Completely different shapes so probably unwise to merge
            continue

        min_shape = [
            min(param.shape[i], model_state.shape[i]) for i in range(len(param.shape))
        ]
        print(name, "model:", mshape, "chkpt:", pshape, "loading:", min_shape)
        idxs = torch.meshgrid(*[torch.arange(s) for s in min_shape])
        model_state[tuple(idxs)].copy_(param[tuple(idxs)])

    return model.load_state_dict(model_state_dict)


def dictdiff(maindict: dict, changeddict: dict) -> dict:
    """
    Returns the difference of two dicts
    """
    assert type(maindict) == dict
    assert type(changeddict) == dict
    from copy import deepcopy

    changeddict = deepcopy(changeddict)

    for k in maindict.keys():
        if k not in changeddict:
            continue

        if type(maindict[k]) == dict:
            # Recursive dict difference had no changes either
            dif = dictdiff(maindict[k], changeddict[k])
            changeddict[k] = dif
            if dif == {}:
                del changeddict[k]
        elif maindict[k] == changeddict[k] or (
            maindict[k] is None and not changeddict[k]
        ):
            del changeddict[k]

    return changeddict


def dict_to_strs(d, prefix="") -> str:
    s = []

    for k, v in d.items():
        if type(v) is dict:
            s += dict_to_strs(v, prefix + k + ".")
        else:
            s += [f"{prefix + k}={v}"]

    return s if prefix != "" else "|".join(s)


def human_readable_time(secs: int) -> str:
    days = secs // (24 * 3600)
    hours = (secs % (24 * 3600)) // 3600
    minutes = (secs % 3600) // 60
    seconds = secs % 60
    return f"{days}d {hours}h {minutes}m {seconds}s"


def human_readable_number(n: int) -> str:
    """Converts long numbers to shorthand:
    1123 -> 1K
    123123 -> 123K
    """
    n = float(n)
    millnames = ["", "K", "M"]
    millidx = max(
        0,
        min(
            len(millnames) - 1, int(math.floor(0 if n == 0 else math.log10(abs(n)) / 3))
        ),
    )

    return "{:.0f}{}".format(n / 10 ** (3 * millidx), millnames[millidx])


def multimap(items: list, func: callable, workers=4, desc=None) -> list:
    """
    Quick and dirty multiprocessing that will return the result of func if it returns None
    """
    from tqdm.contrib.concurrent import process_map

    results = process_map(
        func, items, leave=False, desc=desc, max_workers=workers, total=len(items)
    )
    return list(filter(lambda x: x is not None, results))



import tarfile

from toyds.constants import CACHE_DIR

from os import path, makedirs
from tqdm import tqdm
import os
import subprocess
from urllib import request
import zipfile
import shutil

def dl(link: str):
    if ":" in link:
        return dl_scp_file(link)
    else:
        return dl_http_file(link)


def dl_scp_file(link: str):
    """
    Downloads a file using rsync and scp subprocess

    hostname:runs/nanodrz/nanodrz/1705840799/0013500.pt

    Use rsync --partial --progress --human-readable -e ssh to download the file.
    """
    if ":" not in link:
        raise "Invalid scp path"

    file_path = path.join(CACHE_DIR, link.split(":")[-1])

    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    # Use rsync --partial --progress --human-readable -e ssh
    cmd = [
        "rsync",
        "--partial",
        "--progress",
        "--human-readable",
        "-e",
        "ssh",
        link,
        file_path,
    ]
    print(" ".join(cmd))
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        text=True,
    )

    for line in process.stdout:
        print(line, end="")

    process.wait()

    return file_path


def dl_http_file(link: str):
    file_path = path.join(CACHE_DIR, path.basename(link))

    current_size = 0
    expected_size = int(request.urlopen(link).headers.get("content-length", 0))

    if os.path.exists(file_path):
        current_size = os.path.getsize(file_path)

        if current_size == expected_size:
            return file_path

    headers = {"Range": f"bytes={current_size}-"}
    req = request.Request(link, headers=headers)

    with request.urlopen(req) as response, open(file_path, "ab") as file:
        block_size = 1024 * 1024
        progress_bar = tqdm(
            total=expected_size - current_size,  # Adjust total size
            initial=current_size,
            desc=link,
            unit="B",
            unit_scale=True,
            leave=False,
        )

        while True:
            data = response.read(block_size)
            if not data:
                break
            progress_bar.update(len(data))
            file.write(data)

        progress_bar.close()

    return file_path





import torch
from torch.nn import Module
from torch import Tensor, nn
import torch.nn.functional as F

from .config import Config
from .modules import ScaledSinusoidalEmbedding, Decoder

class GPT(Module):

    def __init__(self, config: Config = Config()):
        super().__init__()

        self.config = config
        modelcfg = config.model
        dmodel = modelcfg.dmodel
        self.num_embs = modelcfg.num_embs

        self.pad_idx = 0
        self.eos_idx = 1
        self.token_emb = nn.Embedding(self.num_embs, dmodel)

        self.token_head = nn.Linear(dmodel, self.num_embs, bias=modelcfg.bias)
        self.init_mod_weights = []
        self.token_pos = ScaledSinusoidalEmbedding(dmodel, max_seqlen=modelcfg.max_seq_len)

        self.decoder = Decoder(
            d_model=dmodel,
            n_heads=modelcfg.nheads,
            n_layers=modelcfg.layers,
            bias=modelcfg.bias,
            dropout=modelcfg.dropout,
        )

        self.init_mod_weights += [
            self.decoder,
            self.token_head,
            self.token_pos,
        ]

        for w in self.init_mod_weights:
            w.apply(self._init_weights)

    def configure_optimizers(
        self, *, weight_decay: float, lr: float, betas: tuple[float, float]
    ):
        # start with all of the candidate parameters
        param_dict = {pn: p for pn, p in self.named_parameters()}
        # filter out those that do not require grad
        param_dict = {pn: p for pn, p in param_dict.items() if p.requires_grad}
        # create optim groups. Any parameters that is 2D will be weight decayed, otherwise no.
        # i.e. all weight tensors in matmuls + embeddings decay, all biases and layernorms don't.
        decay_params = [p for n, p in param_dict.items() if p.dim() >= 2]
        nodecay_params = [p for n, p in param_dict.items() if p.dim() < 2]
        optim_groups = [
            {"params": decay_params, "weight_decay": weight_decay},
            {"params": nodecay_params, "weight_decay": 0.0},
        ]

        optimizer = torch.optim.AdamW(optim_groups, lr=lr, betas=betas, fused=True)

        return optimizer

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
        elif isinstance(module, nn.Conv1d):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)

    def forward(
        self,
        tokens: Tensor,
        loss_funcs: dict[callable],
        lengths: Tensor,
    ):
        max_seq_len = self.config.model.max_seq_len
        pad_tokens  = F.pad(tokens, (0, max(0, max_seq_len - tokens.size(-1))))
        embs = pad_tokens[:, :-1]
        embs = self.token_emb(embs) + self.token_pos(torch.arange(embs.shape[-1]))
        embs = self.decoder(embs)
        embs = self.token_head(embs).permute(0, 2, 1)
        loss = torch.tensor(0, device=tokens.device, dtype=embs.dtype)


        # This is the slightly complicated bit!
        for v in loss_funcs.values():
            loss += v["loss"](embs[v["items"]], tokens, lengths)

        return loss

    @torch.inference_mode()
    def generate(
        self,
        prompt: Tensor,
        temperature=0.8,
        max_steps=10,
        top_k: int = 100,
        top_p: float = 0.0,
    ):
        # TODO
        cfg = self.config.model
        pass

    def from_pretrained(ckpt: str | dict):
        """
        {
            "config": config.model_dump(),
            "step": step,
            "model": {
                k: v
                for k, v in model.state_dict().items()
                if not k.startswith("dac.")
            },
            "optimizer": optimizer.state_dict(),
        }
        """
        if type(ckpt) is str:
            ckpt = torch.load(ckpt)

        config = Config(**ckpt["config"])
        model = GPT(config)
        model.load_state_dict(ckpt["model"], strict=False)
        return model



import torch
import random
import torch.nn.functional as F
from .task import Task

"""
Needle in the haystack problems
"""


def filter_sequence(vocab_size=10, max_len=10):
    """
    Given a sequence, filter out all but a queried set of numbers  in the prompt, producing the numbers from the original sequnce in order that they appear.

    Examples:


    [D, D, C, A, A, A, B, C, D, |, A, |, A, A, A ]

    [D, D, A, B, D, D, C, A, C, |, C, A, | A, C, A, C ]

    """
    pass



def lookup_item(vocab_size=1000, max_seq_len=5000, occurences=1):
    """
    With a large vocab size, determine wether and item is an a sequence

    [A, ......, Z, | Z | T]
    """

    ML = max_seq_len - 4

    # Has to be at least one thing to search for
    minL = 5
    length = random.randint(minL, ML)

    vocab_size -= 1

    randtoken = lambda:  random.randint(4, vocab_size)
    lookfor = randtoken()
    tosearch = torch.randint(4, vocab_size, (length,))

    for i in (tosearch == lookfor).nonzero():
        rand = randtoken()

        while rand == lookfor:
            rand = randtoken()

        tosearch[i] = rand

    contains = random.random() < .5
    if contains:
        idxs = random.sample(list(range(len(tosearch))), k=occurences)
        tosearch[idxs] = lookfor

    # 0 pad 1 EOS 2 Command 3 True 4 False
    contains = 3 if contains else 4
    tosearch = torch.cat((tosearch, torch.tensor([2, lookfor, 2, contains])))
    return tosearch


class LookupItem(Task):

    def generate(self):
        return lookup_item(*self.args, **self.kwargs)

    def train(self, logits, seq, lengths):
        """
        We only really need the last logits
        """
        logits = torch.stack([logits[i, lengths[i]-1] for i in range(len(lengths))])
        targets = torch.stack([seq[i, lengths[i]-1] for i in range(len(lengths))])
        return F.cross_entropy(logits, targets)





import os
from os.path import expanduser as eu


def path_from_env(env: str, default_value: str):
    value = eu(os.getenv(env, default_value))
    os.makedirs(value, exist_ok=True)
    return value

CACHE_DIR = path_from_env("TOYDS_CACHE_DIR", "~/.cache/toyds")
RUN_DIR = path_from_env("TOYDS_RUN_DIR", "~/runs/toyds/")



def reverse():
    """
    Take a sequence and reverse it
    """
    pass


def find_repeat():
    """
    Take a sequence, find a pattern in it and repeat that pattern
    """
    pass


def sort():
    """
    Sort a sequence
    """

def map_one_to_one():
    """
    Given a sequence, learn to map it to another thing

    Mappings can be many to one, one to many
    """


def map_one_to_many():


def map_many_to_one():


def map_many_to_many():
    """
    will need to Define rules for context leading to a result


    """


def numerical():
    """
    Test whether the model can do certain sums
    """


def summarisation():








class Task:
    def __init__(self, *args, **kwargs):
        self.name =self.__class__.__name__.lower()
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        raise NotImplementedError



import os
import time
from contextlib import nullcontext
from functools import partial

import torch
import torch.multiprocessing as mp
from torch.distributed import destroy_process_group, init_process_group
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.nn.utils import clip_grad_norm_
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader

import wandb
from toyds import utils, download
from toyds.config import Config, load_config
from toyds.model import GPT as Model
from toyds.data import ToyDataset
from toyds.tasks.needle import LookupItem
from toyds import optim, download
from toyds.utils import count_parameters, to_device

def train(rank: int, world_size: int, config: Config, dev: bool = False):
    if world_size > 1:
        init_process_group(
            backend="nccl",
            world_size=world_size,
            rank=rank,
        )

    torch.cuda.set_device(rank)
    device = torch.cuda.current_device()

    start_time = time.time()
    train = config.train
    datacfg = config.data

    assert config.run_dir is not None

    is_main_process = rank == 0
    B = train.batch_size

    if is_main_process:
        wandb.init(
            project="toyds",
            config=config.model_dump(),
            name="dev" if dev else None,
        )

    dtype = (
        torch.bfloat16
        if torch.cuda.is_available() and torch.cuda.is_bf16_supported()
        else torch.float16
    )

    model = Model(config).cuda(rank)
    embs = config.model.num_embs
    ds = ToyDataset([LookupItem(vocab_size=embs, max_seq_len=config.model.max_seq_len)])

    def collate_fn(batch):
        sequences = [b[0] for b in batch]
        lengths = torch.tensor([s.shape[-1] for s in sequences])
        loss_funcs = {}
        for i, s in enumerate(batch):
            task = s[1]
            if task.name not in loss_funcs:
                loss_funcs[task.name] = {
                    "loss": task.train,
                    "items": []
                }
            loss_funcs[task.name]["items"].append(i)

        tokens = pad_sequence(sequences, batch_first=True)
        return {"tokens": tokens, "loss_funcs": loss_funcs, "lengths": lengths}

    train_dl = DataLoader(
        ds,
        batch_size=B,
        collate_fn=collate_fn,
        num_workers=datacfg.num_workers,
        pin_memory=True,
        persistent_workers=datacfg.num_workers > 0,
    )

    betas = tuple(train.betas)
    optimizer = model.configure_optimizers(
        weight_decay=train.weight_decay, lr=train.min_lr, betas=betas
    )

    step = 0

    if train.checkpoint is not None:
        checkpoint_path = train.checkpoint

        if "http" in checkpoint_path:
            checkpoint_path = download.dl_http_file(checkpoint_path)

        if ":" in checkpoint_path:
            checkpoint_path = download.dl_scp_file(checkpoint_path)

        checkpoint = torch.load(checkpoint_path, map_location=f"cuda:{rank}")
        utils.load_what_you_can(checkpoint["model"], model)

        if train.continue_from_checkpoint:
            optimizer.load_state_dict(checkpoint["optimizer"])
            step = checkpoint["step"]

        del checkpoint
        torch.cuda.empty_cache()

    # if not dev:
    #     model = torch.compile(model)

    if world_size > 1:
        model = DDP(model, device_ids=[rank])

    if is_main_process:
        print(f"{count_parameters(model) / 1_000_000:.2f}M parameters")

    max_lr = train.max_lr or 10.0 * train.min_lr
    gradient_accumulation_steps = train.grad_acc_steps

    train_dl_iter = iter(train_dl)

    batch = to_device(next(train_dl_iter), device)

    def wandb_log(*args, **kwargs):
        if is_main_process:
            wandb.log(*args, **kwargs)
        if dev:
            print(*args)

    get_lr = getattr(optim, train.lr_schedule)

    wait, warmup, active = 5, 5, 5
    steps = wait + warmup + active if train.torch_profile else train.total_steps

    prof = (
        torch.profiler.profile(
            activities=[
                torch.profiler.ProfilerActivity.CPU,
                torch.profiler.ProfilerActivity.CUDA,
            ],
            schedule=torch.profiler.schedule(
                wait=wait, warmup=warmup, active=active, repeat=1
            ),
            on_trace_ready=torch.profiler.tensorboard_trace_handler("./profiles"),
            record_shapes=False,
            profile_memory=False,
            with_stack=False,  # incurs an additional overhead, disable if not needed
            with_flops=True,
            with_modules=False,  # only for torchscript models atm
        )
        if train.torch_profile
        else nullcontext()
    )

    loss = 0.0
    losses = []
    running = True

    if is_main_process:
        print(f"Took {time.time() - start_time:.2f}s to hit training")

    with prof:
        while step < steps and running:
            t1 = time.perf_counter()
            lr = get_lr(
                step, train.lr_warmup_steps, train.total_steps, train.min_lr, max_lr
            )

            for param_group in optimizer.param_groups:
                param_group["lr"] = lr

            for micro_step in range(gradient_accumulation_steps):
                model.require_backward_grad_sync = (
                    micro_step == gradient_accumulation_steps - 1
                )

                with torch.amp.autocast(
                    enabled=True, device_type="cuda", dtype=dtype
                ), torch.backends.cuda.sdp_kernel(**train.flash.model_dump()):
                    loss = model(**batch)
                    loss = loss / gradient_accumulation_steps

                if torch.isnan(loss):
                    wandb.alert("Nan Detection", "Stopping!")
                    running = False
                    continue

                loss.backward()
                batch = to_device(next(train_dl_iter), device)

            if is_main_process and step == 0:
                print(
                    f"Took {time.time() - start_time:.2f}s to finish first step training loop"
                )

            grad_norm = clip_grad_norm_(model.parameters(), train.max_grad_norm)

            optimizer.step()
            optimizer.zero_grad()
            step += 1

            t2 = time.perf_counter()

            if train.torch_profile:
                prof.step()

            if step % train.log_every == 0:
                loss = loss.item() * gradient_accumulation_steps
                # Determining loss slope for explosion detection / divergence
                # losses.append(loss)
                # loss_slope = optim.calculate_smoothed_slope(
                #     losses,
                #     regression_win=train.regression_win,
                #     smoothing_constant=train.regression_smoothing,
                # )

                # if loss_slope > 1e-3 and steps > 400:
                #     wandb.alert("Explosion Warning", "Check loss graph")

                metrics = {
                    "train/loss": loss,
                    "train/grad_norm": grad_norm.item(),
                    "train/lr": lr,
                    "train/batch_duration": t2 - t1,
                    # "train/loss_slope": loss_slope,
                }

                wandb_log(
                    metrics,
                    step=step,
                )

            if step % train.checkpoint_every == 0 and is_main_process:
                checkpoint = {
                    "config": config.model_dump(),
                    "step": step,
                    "model": {
                        k: v
                        for k, v in model.state_dict().items()
                        if not k.startswith("dac.")
                    },
                    "optimizer": optimizer.state_dict(),
                }
                file_url = os.path.join(config.run_dir, f"{step:07}.pt")
                print(f"Saved checkpoint to {file_url}")
                torch.save(checkpoint, file_url)

    if world_size > 1:
        destroy_process_group()


def main(
    config: str | Config,
    dev: bool = False,
    profile: bool = False,
):
    config: Config = load_config(config)

    config.train.torch_profile = config.train.torch_profile or profile

    if dev:
        print("Running in dev mode (smaller dataset, batch size, fewer epochs, etc.)")
        config.data.num_workers = 0

    if config.train.is_resuming:
        print(f"Resuming from {config.train.checkpoint}. Incrementing seed.")
        # TODO save the seed state in the checkpoint
        config.seed += 1

    if "MASTER_ADDR" not in os.environ:
        os.environ["MASTER_ADDR"] = "127.0.0.1"

    if "MASTER_PORT" not in os.environ:
        os.environ["MASTER_PORT"] = "17778"

    world_size = config.train.gpus
    if world_size is None:
        world_size = torch.cuda.device_count()

    if world_size > 1:
        mp.spawn(
            train,
            args=(world_size, config, dev),
            nprocs=world_size,
            join=True,
        )
    else:
        train(0, world_size, config, dev)

def cli_main():
    import argparse

    parser = argparse.ArgumentParser(description="GPT Training")

    # Add a positional argument 'config' which expects a string input
    parser.add_argument("config", help="Path to configuration file")

    # Add optional boolean flags with short and long versions
    parser.add_argument("-d", "--dev", action='store_true', help="Run in dev mode")
    parser.add_argument("-p", "--profile", action='store_true', help="Profile memory usage")

    args = parser.parse_args()

    # Access values from parsed arguments
    config = args.config
    dev = args.dev
    profile = args.profile
    main(config, dev, profile)


if __name__ =="__main__":
    cli_main()



import math
import torch


def warmup_then_linear_decay(
    step: int, warmup_steps: int, total_steps: int, min_lr: float, max_lr: float
) -> float:
    if step <= warmup_steps:
        return min_lr + step * (max_lr - min_lr) / (warmup_steps)
    elif step > total_steps:
        return min_lr
    else:
        return max_lr - max_lr / (total_steps - warmup_steps) * (step - warmup_steps)


def warmup_then_cosine_decay(
    step: int, warmup_steps: int, total_steps: int, min_lr: float, max_lr: float
):
    if step < warmup_steps:
        return min_lr + step * (max_lr - min_lr) / (warmup_steps)
    elif step > total_steps:
        return min_lr
    else:
        decay_ratio = (step - warmup_steps) / (total_steps - warmup_steps)
        assert 0 <= decay_ratio <= 1
        coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))  # coeff ranges 0..1
        return min_lr + coeff * (max_lr - min_lr)


def warmup_then_inv_sqrt_decay(
    step: int, warmup_steps: int, total_steps: int, min_lr: float, max_lr: float
):
    if step < warmup_steps:
        return min_lr + step * (max_lr - min_lr) / (warmup_steps)
    else:
        decay_factor = max_lr * warmup_steps**0.5
        return decay_factor * step**-0.5


def warmup_then_constant(
    step: int, warmup_steps: int, total_steps: int, min_lr: float, max_lr: float
):
    if step < warmup_steps:
        return min_lr + step * (max_lr - min_lr) / (warmup_steps)
    else:
        return max_lr


@torch.jit.script
def calculate_smoothed_slope(
    losses: list[float],
    smooth_window_size: int = 3,
    smoothing_constant: float = 0.93,
    regression_win: int = 20,
) -> float:
    y = torch.tensor(losses[-smooth_window_size - 10 :]).view(-1, 1)
    x = torch.arange(y.shape[0]).view(-1, 1)

    # Smooth the data using a simple moving average with a smoothing constant
    smooth_window_size = 3
    conv_weights = (smoothing_constant / smooth_window_size) * torch.ones(
        smooth_window_size
    )
    smoothed_y = torch.nn.functional.conv1d(
        y.view(1, 1, -1),
        conv_weights.view(1, 1, -1),
        padding=(smooth_window_size - 1) // 2,
    ).view(-1)

    regression_win = min(regression_win, smoothed_y.shape[-1])
    # Perform linear regression on the smoothed_y
    x_last_n = x[-regression_win:]
    y_last_n = smoothed_y[-regression_win:]
    X = torch.cat([torch.ones(regression_win, 1), x_last_n], dim=1)
    coefficients = torch.linalg.lstsq(X, y_last_n.view(-1, 1)).solution.flatten()
    return coefficients[1]
